/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Soup;
using Diorite.Logger;
using Diorite.System;


/* There are changes in VAPI between Vala 0.14 a 0.16 */
private extern void soup_session_add_feature(Soup.Session session, Soup.SessionFeature feature);
private extern void soup_session_add_feature_by_type(Soup.Session session, GLib.Type feature_type);
//~ private extern void soup_session_remove_feature(Soup.Session session, Soup.SessionFeature feature);
private extern void soup_session_remove_feature_by_type(Soup.Session session, GLib.Type feature_type);

namespace Diorite{
	
	
	
	private const string TSOCKS_CONF = "tsocks.conf";
	private const string COOKIES = "cookies.dat";
	private int LIBSOUP_VERSION[3];
	private string LIBSOUP_VERSION_STRING;
	
	/**
	 * Proxy settings
	 * 
	 * AUTO = http_proxy and https_proxy environment variables are used to resolve proxy
	 * settings; MANUAL = manual proxy settings is used (type, server and port);
	 * NONE = direct network connection is used
	 */
	public enum ProxySettings{
		AUTO,
		MANUAL,
		NONE;
	}
	
	/**
	 * Type of proxy server
	 * 
	 * HTTP = HTTP(S) proxy server; SOCKS = SOCKS4 or SOCKS5 proxy server
	 */
	public enum ProxyType{
		HTTP, SOCKS;
	}
	
	public errordomain ProxyError{
		SOCKS_NOT_AVAILABLE,
		NO_PROXY_SERVER,
		NO_PROXY_TYPE,
		NO_PROXY_PORT
	}
	
	
	/**
	 * Network connection and related settings including
	 * proxy settings, cookies and cache
	 */
	public class Connection: GLib.Object{
		private Storage storage;
		public Soup.Session session {get; private set;}
		private bool proxy_is_set = false;
		
		/**
		 * Construct new connection manager
		 * 
		 * @param session session used for connections
		 * @param storage storage for cookies and cache
		 */
		public Connection.with_session(Soup.Session session, Storage storage, Soup.LoggerLogLevel logging=Soup.LoggerLogLevel.NONE){
			parse_libsoup_version();
			assert(LIBSOUP_VERSION[0] > 2 || (LIBSOUP_VERSION[0] == 2 && LIBSOUP_VERSION[1] >= 38));
			this.storage = storage;
			this.session = session;
			if(logging != Soup.LoggerLogLevel.NONE){
				var logger = new Soup.Logger(logging, 1024);
				soup_session_add_feature(session, logger);
			}
		}
		
		~Connection()
		{
			dump_cache();
		}
		
		/**
		 * Setup network proxy
		 * 
		 * @param mode Mode used for proxy settings
		 * @param type Type of proxy server. Used only with manual mode.
		 * @param server Address of proxy server. Used only with manual mode.
		 * @param port Port of proxy server. Used only with manual mode.
		 * @throws ProxyError on failure
		 **/
		public void set_up_proxy(ProxySettings mode = ProxySettings.AUTO,
			ProxyType type, string? server = null, string? port = null)
			throws ProxyError{
			
			// Required arguments for manual settings
			if(mode == ProxySettings.MANUAL){
				if(server == null){
					throw new ProxyError.NO_PROXY_SERVER("Proxy server is not specified");
				}
				if(port == null){
					throw new ProxyError.NO_PROXY_PORT("Proxy port is not specified");
				}
			}
			
			
			/* Attempt to change settings */
			if(proxy_is_set){
				var tsocks_conf_file = this.storage.get_config_path(TSOCKS_CONF);
				// Setup wrapper around tsocks library and return
				if(mode == ProxySettings.MANUAL && type == ProxyType.SOCKS){
					Logger.lib_debug("Setting proxy (manual): socks://%s:%s", server, port);
					var conf = "server = %s\nserver_port = %s\n".printf(server, port);
					
					try{
						overwrite_file(tsocks_conf_file, conf);
					}
					catch(Error e){
						throw new ProxyError.SOCKS_NOT_AVAILABLE(
							"Unable to store SOCKS config file: %s".printf(e.message)
						);
					}
					
					if(!have_socks()){
						throw new ProxyError.SOCKS_NOT_AVAILABLE(
							"Unable to initialize SOCKS proxy wrapper"
						);
					}
				}
				else{
					if(tsocks_conf_file.query_exists()){
						try{
							tsocks_conf_file.delete();
						}
						catch(Error e){
							Logger.lib_warning("Unable to discard SOCKS settings");
						}
					}
				}
			
				/* Don't change HTTPS proxy settings.
				 * TODO: libsoup used to crash when proxy setting has been
				 * changed in run time. Does the issue persist? */
				return;
			}
			
			// Unset proxy
			session.abort();
			soup_session_remove_feature_by_type(session, typeof(Soup.ProxyURIResolver));
			soup_session_remove_feature_by_type(session, typeof(Soup.ProxyResolverDefault));
			session.proxy_uri = null;
			
			// HTTPS proxy is supported directly in Soup library
			string? proxy = null;
			if(mode == ProxySettings.MANUAL){
				if(type != ProxyType.SOCKS){
					proxy = "http://%s:%s".printf(server, port);
					Logger.lib_debug("Setting proxy (manual): %s", proxy);
				}
			}
			else if(mode == ProxySettings.AUTO){
				// ProxyResolverDefault is powerful, it supports also PAC
				Logger.lib_debug("Setting proxy (auto): %s", "dynamic resolver");
				soup_session_add_feature_by_type(session, typeof(Soup.ProxyResolverDefault));
				
			}
			else{
				Logger.lib_debug("No proxy settings applied.");
			}
			
			
			if(proxy != null && proxy != ""){
				session.proxy_uri = new Soup.URI(proxy);
			}
			
			proxy_is_set = true;
		}
		
		public static bool have_socks(out string? error=null){
			int result = -1;
			error = null;
			try{
				string[] args = {"which", "tsocks"};
				Process.spawn_sync(null, args, null,
				 SpawnFlags.SEARCH_PATH|SpawnFlags.STDOUT_TO_DEV_NULL|SpawnFlags.STDERR_TO_DEV_NULL,
				 null, null, null, out result);
			}
			catch(SpawnError e){
				lib_warning("SpawnError: %s", e.message);
				return false;
			}
			return result == 0;
		}
		
		/**
		 * Set up cookies
		 * 
		 * @param clear also clears cookies storage if true
		 */
		public void set_up_cookies(bool clear=false){
			var file = this.storage.get_config_path(COOKIES);
			if(clear && file.query_exists()){
				try{
					file.delete();
				}
				catch(Error e){
					Logger.lib_warning("Unable to delete cookies file: %s", file.get_path());
				}
			}
			
			weak SessionFeature? feature = this.session.get_feature(typeof(CookieJarText));
			if(feature != null) this.session.remove_feature(feature);
			soup_session_add_feature(session, new CookieJarText(file.get_path(), false));
		}
		
		/**
		 * Enable cache for HTTP objects and set its size
		 * 
		 * @param max_size maximal size of cache in MB
		 * @return true on success, false if cache is already enabled
		 */
		public bool enable_cache(int max_size){
			lib_debug("Max data cache size: %d", max_size);
			Cache? cache = this.session.get_feature(typeof(Cache)) as Cache;
			if(cache != null){
				cache.set_max_size(max_size * 1024 * 1024);
				return false;
			}
			
			var dir = storage.get_cache_path("browser");
			cache = new Cache(dir.get_path(), CacheType.SINGLE_USER);
			cache.set_max_size(max_size * 1024 * 1024);
			cache.load();
			soup_session_add_feature(session, (owned) cache);
			return true;
		}
		
		/**
		 * Disable cache for HTTP objects
		 * 
		 * @return true on succes, false when cache is not enabled
		 */
		public bool disable_cache(){
			Cache? cache = this.session.get_feature(typeof(Cache)) as Cache;
			if(cache == null) return false;
			
			session.remove_feature(cache);
			return true;
		}
		
		/**
		 * Clear cache for HTTP objects
		 * 
		 * @return true on succes, false when cache is not enabled
		 */
		public bool clear_cache(){
			Cache? cache = this.session.get_feature(typeof(Cache)) as Cache;
			if(cache == null) return false;
			
			cache.clear();
			cache.flush();
			return true;
		}
		
		/**
		 * Dumps cache to disk
		 */
		public void dump_cache(){
			SessionFeature? feature = this.session.get_feature(typeof(Cache));
			if(feature == null) return;
			weak Cache cache = feature as Cache;
			cache.dump();
		}
		
		/**
		 * Downloads remote file
		 * 
		 * @param uri		URI of remote file
		 * @param File		local file data will be stored to
		 * @param status	response status code
		 * @return 			true on success, false on failure
		 */
		public bool download_file(string uri, File local_file, out uint? status=null){
			var msg = new Soup.Message("GET", uri);
			var message_status = session.send_message(msg);
			status = message_status;
			if(message_status != 200){
				return false;
			}
			
			unowned Soup.MessageBody body = msg.response_body;
			var dir = local_file.get_parent();
			if(!dir.query_exists(null)){
				try{
					dir.make_directory_with_parents(null);
				}
				catch(GLib.Error e){
					lib_critical("Unable to create directory: %s", e.message);
				}
			}
			
			FileOutputStream stream;
			try{
				stream = local_file.replace(null, false, FileCreateFlags.REPLACE_DESTINATION, null);
			}
			catch(GLib.Error e){
				lib_critical("Unable to create local file: %s", e.message);
				return false;
			}
			
			try{
				stream.write_all(body.data, null, null);
			}
			catch(IOError e){
				lib_critical("Unable to store remote file: %s", e.message);
				return false;
			}
			try{
				stream.close();
			}
			catch(IOError e){
				lib_warning("Unable to close stream: %s", e.message);
			}
			
			return true;
		}
		
		private static void parse_libsoup_version(){
			// do not use if(LIBSOUP_VERSION != null),
			// because the fixed-size array has been allocated
			if(LIBSOUP_VERSION[0] != 0) return;
			LIBSOUP_VERSION_STRING = Config.LIBSOUP_2_4_VERSION;
			var libsoup_version = LIBSOUP_VERSION_STRING.split("."); 
			LIBSOUP_VERSION = {
				int.parse(libsoup_version[0]),
				int.parse(libsoup_version[1]),
				int.parse(libsoup_version[2])
			};
			lib_debug("Libsoup version: %d.%d.%d",
				LIBSOUP_VERSION[0],
				LIBSOUP_VERSION[1],
				LIBSOUP_VERSION[2]);
		}
		
		public static string libsoup_version_string()
		{
			parse_libsoup_version();
			return LIBSOUP_VERSION_STRING;
		}
	}
}
