/*
 * Copyright 2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


using Diorite.Logger;

namespace Nuvola{

const string EXTENSIONS_KEY = "extensions";
/**
 * Extension Manager is responsible for loading and unloading extensions.
 * 
 * Extensions are currently kind of fake extensions, because they are included
 * in the main application binary (nuvolaplayer), but they will be build
 * as separate libraries (libnuvolaplayer-extensionid.so) in the future
 */
public class ExtensionsManager{
	private weak Gtk.UIManager ui_manager;
	private weak MainWindow window;
	private HashTable<string, Extension> active_extensions;
	public HashTable<string, ExtensionInfo?> available_extensions { get; private set;}
	private ObjectContainer objects;
	
	/**
	 * Creates new Extensions Manager
	 * 
	 * @param ui_manager	main UI manager
	 * @param window		main application window
	 * @param objects		object container providing all necessary dependencies for extensions
	 */
	public ExtensionsManager(Gtk.UIManager ui_manager, MainWindow window, ObjectContainer objects){
		this.objects = objects;
		this.ui_manager = ui_manager;
		this.window = window;
		active_extensions = new HashTable<string, Extension>(str_hash, str_equal);
		available_extensions = new HashTable<string, ExtensionInfo?>(str_hash, str_equal);
		
		window.add_actions_ui.connect(on_add_actions_ui);
		window.remove_actions_ui.connect(on_remove_actions_ui);
		find_extensions();
	}
	
	/**
	 * Returns extension by id
	 * 
	 * @param id extension id
	 * @return extension instance or null if not found
	 */
	public unowned Extension? get(string id){
		return active_extensions.lookup(id);
	}
	
	/**
	 * Loads extension instance
	 * @param id id of the extension
	 * @return	extension or null if not found
	 */
	public unowned Extension? load(string id){
		weak Extension? result = get(id);
		if(result != null){
			return result;
		}
		
		ExtensionInfo? info = available_extensions.lookup(id);
		if(info != null){
			var extension = Object.new(info.type, "id", id) as Extension;
			if(extension != null){
				try{
					extension.load(objects);
					message("Extension with id '%s' loaded.", id);
				}
				catch(Error e){
					warning("Unable to load extension '%s': %s", id, e.message);
					return null;
				}
				
				extension.add_actions_ui(ui_manager);
				active_extensions.insert(id, extension);
				result = extension;
				return result;
			}
			
		}
		return null;
	}
	
	/**
	 * Uloads extension with id id
	 */
	public bool unload(string id){
		var extension = get(id);
		if(extension != null){
			extension.remove_actions_ui(ui_manager);
			extension.unload();
			active_extensions.remove(id);
			message("Extension with id '%s' unloaded.", id);
			return true;
		}
		return false;
	}
	
	/**
	 * Unloads all extensions.
	 */
	public void unload_all(){
		foreach(string extension in active_extensions.get_keys()){
			unload(extension);
		}
	}
	
	/**
	 * Called when main window requires ui definitions for actions.
	 */
	private void on_add_actions_ui(Gtk.UIManager manager){
		foreach(var extension in active_extensions.get_values()){
			extension.add_actions_ui(manager);
		}
	}
	
	/**
	 * Called when main window requires removal of ui definitions for actions.
	 */
	private void on_remove_actions_ui(Gtk.UIManager manager){
		foreach(var extension in active_extensions.get_values()){
			extension.remove_actions_ui(manager);
		}
	}
	
	private void find_extensions(){
		// TODO: load extensions from libraries
		
		// Built-in extensions
		available_extensions.insert("trayicon", Extensions.TrayIcon.get_info());
		#if LASTFM
			available_extensions.insert("lastfm", Extensions.Lastfm.get_info());
		#endif
		available_extensions.insert("lyrics", Extensions.Lyrics.get_info());
		available_extensions.insert("mediakeys", Extensions.MediaKeys.get_info());
		#if NOTIFICATIONS
			available_extensions.insert("notifications", Extensions.Notifications.get_info());
		#endif
		available_extensions.insert("mpris", Extensions.Mpris.get_info());
		available_extensions.insert("dockmanager", Extensions.DockManager.get_info());
		#if UNITY_QUICK_LIST
			available_extensions.insert("unityquicklist", Extensions.UnityQuickList.get_info());
		#endif
		available_extensions.insert("developerbar", Extensions.DeveloperBar.get_info());
		#if EXPERIMENTAL
			available_extensions.insert("sample", Extensions.Sample.get_info());
		#endif
	}
}

/**
 * Struct holding extension information
 */
public struct ExtensionInfo{
	/**
	 * Human-readable name of the extension
	 */
	public string name;
	/**
	 * Version of the extension
	 */
	public string version;
	/**
	 * Description of the extension
	 */
	public string description;
	/**
	 * Author of the extension
	 */
	public string author;
	/**
	 * GLib type of the extension class
	 */
	public Type type;
	/**
	 * Load the extension automatically unless disabled.
	 */
	public bool autoload;
	
	/**
	 * Creates new struct with extension info
	 */
	public ExtensionInfo(string name, string description, string author, Type type, bool autoload=true){
		this.name = name;
		this.description = description;
		this.author = author;
		this.type = type;
		this.autoload = autoload;
	}
}

/**
 * Errors related to extensions.
 */
public errordomain ExtensionError{
	/**
	 * Extension requires object that doesn't exist
	 */
	OBJECT_NOT_FOUND,
	/**
	 * Extension fails to load
	 */
	EXTENSION_FAILURE;
}

} // namespace Nuvola
