/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Nuvola.Settings;
using Diorite;

namespace Nuvola
{

namespace Actions{
	public const string PURGE_COOKIES = "purge-cookies";
	public const string PURGE_CACHE = "purge-cache";
	public const string PURGE_SETTINGS = "purge-settings";
	public const string PURGE_UI_STATE = "purge-ui-state";
	public const string DISABLE_PROXY = "disable-proxy";
}

namespace Settings
{
	// Configuration keys
	public const string PROXY_TYPE = "proxy_type";
	public const string PROXY_SERVER = "proxy_server";
	public const string PROXY_PORT = "proxy_port";
	public const string PROXY_SETTINGS = "proxy_settings";
	public const string HIDE_WINDOW = "hide_window";
	public const string PAGE_CACHE = "page_cache";
	public const string DATA_CACHE = "data_cache";
	public const string DATA_CACHE_SIZE = "data_cache_size";
	public const string WEB_PLUGINS_BLOCKER = "web_plugins";
	public const string SPACE_KEY = "space_key";
	public const string BLOCK_INCOMPATIBLE_FLASH = "block_incompatible_flash";
	public const string DARK_THEME = "dark_theme";
	public const string UPDATES_SERVER = "updates_server";
	public const string CHECK_UPDATES = "check_updates";
}


/**
 * Interface providing access to configuration options via GObject
 * properties with right type and right default value.
 * 
 * This implementation uses configuration file as a storage.
 */
public class Configuration: MapFile
{
	/**
	 * Creates new configuration storage using specified file
	 * 
	 * @param file file used to store configuration
	 */
	public Configuration(File file)
	{
		base(file, false);
	}
	
	[Description(nick = "Proxy settings", blurb = "Defines what proxy settings is used: no, system or manual settings.")]
	public ProxySettings proxy_settings
	{
		get
		{
			var val = get_int(PROXY_SETTINGS, ProxySettings.AUTO,
				ProxySettings.AUTO, ProxySettings.NONE);
			return (ProxySettings) val;
		}
		set
		{
			set_int(PROXY_SETTINGS, (int) value);
		}
	}
	[Description(nick = "Proxy type", blurb = "Defines what proxy type is used: HTTP(S) or SOCKS 4/5")]
	public ProxyType proxy_type
	{
		get
		{
			var val = get_int(PROXY_TYPE, ProxyType.HTTP,
				ProxyType.HTTP, ProxyType.SOCKS);
			return (ProxyType) val;
		}
		set
		{
			set_int(PROXY_TYPE, (int) value);
		}
	}
	[Description(nick = "Proxy server", blurb = "Address of proxy server.")]
	public string proxy_server
	{
		owned get
		{
			return this[PROXY_SERVER] ?? "";
		}
		set
		{
			this[PROXY_SERVER] = value ?? "";
		}
	}
	[Description(nick = "Proxy server port", blurb = "Port of proxy server.")]
	public string proxy_port
	{
		owned get
		{
			return this[PROXY_PORT] ?? "";
		}
		set
		{
			this[PROXY_PORT] = value ?? "";
		}
	}
	[Description(nick = "Window hiding", blurb = "Defines whether hide window when close button is pressed.")]
	public OptionState hide_window
	{
		get
		{
			return get_state(HIDE_WINDOW);
		}
		set
		{
			set_state(HIDE_WINDOW, value);
		}
	}
	[Description(nick = "Web plugins", blurb = "Whether web plugins are enabled.")]
	public bool web_plugins_blocker
	{
		get
		{
			return !get_bool(WEB_PLUGINS_BLOCKER, false);
		}
		set
		{
			set_bool(WEB_PLUGINS_BLOCKER, !value);
		}
	}
	[Description(nick = "Size of data cache", blurb = "Defines how much data cache in MB is used to cache remote objects.")]
	public int data_cache_size
	{
		get
		{
			return get_int(DATA_CACHE_SIZE, 100, 0, 1024);;
		}
		set
		{
			set_int(DATA_CACHE_SIZE, value);
		}
	}
	[Description(nick = "Data cache", blurb = "Enables or disables cache for remote data (images, scripts, stylesheets, etc.).")]
	public bool data_cache
	{
		get
		{
			return this.get_bool(DATA_CACHE, true);
		}
		set
		{
			this.set_bool(DATA_CACHE, value);
		}
	}
	[Description(nick = "Use space key", blurb = "Whether it is allowed to use space key in keybindings.")]
	public bool space_key
	{
		get
		{
			return get_bool(SPACE_KEY, true);
		}
		set
		{
			set_bool(SPACE_KEY, value);
		}
	}
	
	[Description(nick = "Block incompatible flash", blurb = "Whether to block incompatible Flash plugin.")]
	public bool block_incompatible_flash
	{
		get
		{
			return get_bool(BLOCK_INCOMPATIBLE_FLASH, true);
		}
		set
		{
			set_bool(BLOCK_INCOMPATIBLE_FLASH, value);
		}
	}
	
	[Description(nick = "Use dark theme", blurb = "Use dark GTK+ theme.")]
	public bool dark_theme
	{
		get
		{
			return get_bool(DARK_THEME, false);
		}
		set
		{
			set_bool(DARK_THEME, value);
		}
	}
	
	public string updates_server
	{
		owned get
		{
			var server =  this[UPDATES_SERVER] ?? "";
			return server != "" ? server : "https://nuvolaplayer.fenryxo.cz/service_updates_service";
			// python3 -m http.server 1091
		}
		set
		{
			this[UPDATES_SERVER] = value ?? "";
		}
	}
	
	public bool check_updates
	{
		get
		{
			return get_bool(CHECK_UPDATES, true);
		}
		set
		{
			set_bool(CHECK_UPDATES, value);
		}
	}
}

} // namespace Nuvola
