/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Nuvola
{

/**
 * WebBackend uses WebKit web rendering engine and JavaScriptCore script engine
 * to load web streaming services providing music playback.
 */
public class WebBackend: GLib.Object
{
	private static const string SERVICE_LAST_URI = "last_uri";
	private static const string SERVICE_ZOOM_LEVEL = "zoom_level";
	private static const string SERVICE_GSTREAMER = "gstreamer";
	private static const string SERVICE_FLASH = "flash_enabled";
	
	[Description(nick = "Web plugins", blurb = "Whether web plugins are enabled.")]
	public bool web_plugins { get; set; }
	
	[Description(nick = "Web plugins blocker", blurb = "Whether web plugins blocker is enabled.")]
	public bool web_plugins_blocker { get; set; }
	
	[Description(nick = "Streaming service", blurb = "Currently loaded streaming service.")]
	public Service? service { get; private set; }
	
	[Description(nick = "JavaScript API", blurb = "JavaScript API to control the service.")]
	public JSApi api { get; private set; }
	
	private Diorite.Application app;
	private Diorite.Storage storage;
	private MainWebView view;
	private Diorite.Actions? actions;
	private Diorite.MapFile ui_state;
	private Configuration config;
	private FormatSupport formats;
	private SList<string> blocked_resources;
	private HashTable<string, string> libs;
	
	/**
	 * Creates new WebBackend
	 * 
	 * @param storage    storage used to load scripts
	 * @param view       web view to load streaming services in
	 * @param ui_state   user interface state settings
	 * @param config     configuration
	 * @param actions    actions that can be triggered by scripts
	 */
	public WebBackend(Diorite.Application app, Diorite.Storage storage, MainWebView view, Configuration config,
	Diorite.MapFile ui_state, FormatSupport formats, Diorite.Actions? actions)
	{
		this.app = app;
		this.storage = storage;
		this.view = view;
		this.actions = actions;
		this.ui_state = ui_state;
		this.config = config;
		this.formats = formats;
		blocked_resources = new SList<string>();
		view.settings.enable_default_context_menu = true;
		view.notify["zoom-level"].connect(on_zoom_level_changed);
		view.create_plugin_widget.connect(on_create_plugin_widget);
		view.resource_request_starting.connect(on_resource_request);
		view.mime_type_policy_decision_requested.connect(on_mime_type_policy_decision_requested);
		view.download_requested.connect(on_download_requested);
		view.actions_enabled = false;
		
		libs = new HashTable<string, string>(str_hash, str_equal);
		uint major;
		uint minor;
		uint micro;
		uint nano;
		Gst.version(out major, out minor, out micro, out nano);
		libs.insert("gstreamer", "%u.%u.%u".printf(major, minor, micro));
		libs.insert("gstreamer_major_minor", "%u.%u".printf(major, minor));
		libs.insert("soup", Diorite.Connection.libsoup_version_string());
		libs.insert("webkitgtk", "%u.%u.%u".printf(WebKit.major_version(), WebKit.minor_version(), WebKit.micro_version()));
		libs.insert("webkitgtk_major_minor", "%u.%u".printf(WebKit.major_version(), WebKit.minor_version()));
	}
	
	public signal void download_started(WebKit.Download download, File destination);
	
	/**
	 * Loads service and creates JS API for it
	 * 
	 * @param service    service to load or null to unload current service
	 */
	public void load_service(Service? service)
	{
		this.service = service;
		if (api != null)
		{
			api.user_warning.disconnect(on_user_warning);
			api.user_question.disconnect(on_user_question);
			api.deactivate();
		}
		
		api = service != null ? new JSApi(storage, view, service, actions, JSApi.MAIN_FRAME_ID, formats, libs) : null;
		
		if (service == null)
		{
			view.service = null;
			view.notify["uri"].disconnect(on_uri_changed);
			view.actions_enabled = false;
			view.open("about:blank");
			view.notify["uri"].connect_after(on_uri_changed);
		}
		else
		{
			api.user_warning.connect(on_user_warning);
			api.user_question.connect(on_user_question);
			view.service = service;
			view.actions_enabled = true;
			var settings =  view.get_settings();
			
			var flash_enabled = web_plugins && config.get_bool(ServicesManager.config_key(service, SERVICE_FLASH), service.flash_plugin);
			formats.flash_enabled = flash_enabled;
			settings.enable_plugins = flash_enabled;
			view.set_settings(settings);
			var uri = ui_state[ServicesManager.config_key(service, SERVICE_LAST_URI)] ?? service.home_page;
			debug("Starting web interface for %s: %s", service.id, uri);
			view.scaled_zoom_level = (float) ui_state.get_double(ServicesManager.config_key(service, SERVICE_ZOOM_LEVEL), 1.0);
			view.notify["uri"].disconnect(on_uri_changed);
			view.open(uri);
			view.notify["uri"].connect_after(on_uri_changed);
			var gst_key = "gstreamer_" + libs.lookup("gstreamer_major_minor");
			if (!config.get_bool(gst_key, true))
				formats.disable_gstreamer();
			
			
		}
	}
	
	/**
	 * Get widget for service's settings
	 * 
	 * @return a widget if service has settings, null otherwise
	 */
	public Gtk.Widget? get_settings()
	{
		var grid = new Gtk.Grid();
		
		var label2 = new Diorite.WrappedLabel(_("Load Flash plugin"));
		label2.hexpand = false;
		var toggle2 = new Gtk.CheckButton();
		toggle2.add(label2);
		toggle2.expand = false;
		toggle2.active = formats.flash_enabled;
		toggle2.notify["active"].connect_after(on_flash_toggled);
		grid.attach(toggle2, 0, 1, 1, 1);
		
		if (service != null && JSApi.has_settings(service))
		{
			
			var service_view = new StaticWebView();
			var js_api = new JSApi(storage, service_view, service, actions, JSApi.PREFERENCES_FRAME_ID, formats, libs);
			js_api.config_changed.connect(on_service_config_changed);
			
			var template_file = storage.get_data_file("html/Settings.html");
			service_view.static_uri = template_file.get_uri();
			service_view.load_uri(service_view.static_uri);
			service_view .margin_top = 15;
			var service_scroll = new Gtk.ScrolledWindow(null, null);
			service_scroll.add(service_view);
			service_scroll.show_all();
			service_scroll.expand = true;
			grid.attach(service_scroll, 0, 2, 1, 1);
		}
		grid.show_all();
		return grid;
	}
	
	public unowned HashTable<string, string> get_libs()
	{
		return libs;
	}
	
	public void disable_gstreamer()
	{
		formats.disable_gstreamer();
		view.reload();
	}
	
	public void enable_gstreamer()
	{
		app.show_user_info("Restart needed to re-enable HTML5 Audio", "It is necessary to restart Nuvola Player to enable GStreamer/HTML5 Audio support. This cannot be done in run time.");
	}
	
	private void on_zoom_level_changed()
	{
		if (service == null)
			return;
		ui_state.set_double(ServicesManager.config_key(service, SERVICE_ZOOM_LEVEL),
			view.scaled_zoom_level);
	}
	
	private void on_uri_changed(GLib.Object o, ParamSpec p)
	{
		return_if_fail(_service != null);
		ui_state[ServicesManager.config_key(service, SERVICE_LAST_URI)] = view.uri;
	}
	
	private void on_service_config_changed()
	{
		return_if_fail(api != null);
		try
		{
			api.send_message(JSApi.CONFIG_CHANGED_MESSAGE);
		}
		catch (JSError e)
		{
			/* Service integration doesn't handle config-changed signal.
			 * No big deal. */
		}
	}
	
	private unowned Gtk.Widget on_create_plugin_widget(string mime, string uri, GLib.HashTable attributes)
	{
		if (web_plugins_blocker && service != null && uri != null && !service.matches_sandbox(uri))
		{
			debug("Widget blocked: %s", uri);
			blocked_resources.append(Uri.unescape_string(uri));
		}
		else
			debug("Widget allowed: %s", uri);
		
		unowned Gtk.Widget widget = null; // HACK
		return widget;
	}
	
	private void on_resource_request(WebKit.WebFrame frame, WebKit.WebResource resource,
	WebKit.NetworkRequest request, WebKit.NetworkResponse? response)
	{
		var uri = request.uri;
		if (uri == null || !web_plugins_blocker)
			return;
		uri = Uri.unescape_string(uri);
		unowned SList<string> item = blocked_resources.find_custom(uri, strcmp);
		if (item != null)
		{
			request.uri = "about:blank";
			blocked_resources.delete_link(item);
		}
	}
	
	private void on_flash_toggled(GLib.Object o, ParamSpec p)
	{
		var flash_enabled = (o as Gtk.CheckButton).active;
		config.set_bool(ServicesManager.config_key(service, SERVICE_FLASH), flash_enabled);
		var settings =  view.get_settings();
		formats.flash_enabled = flash_enabled;
		settings.enable_plugins = flash_enabled;
		view.set_settings(settings);
		view.reload();
	}
	
	private bool on_mime_type_policy_decision_requested(WebKit.WebFrame frame, WebKit.NetworkRequest request, string mimetype, WebKit.WebPolicyDecision decision)
	{
		if (!view.can_show_mime_type(mimetype))
		{
			decision.download();
			return true;
		}
		return false;
	}
	
	private bool on_download_requested(GLib.Object download)
	{
		var request = download as WebKit.Download;
		return_val_if_fail(request != null, false);
		var dir = GLib.Environment.get_user_special_dir(UserDirectory.DOWNLOAD);
		if (dir == null)
			dir = GLib.Environment.get_home_dir();
		var filename = request.suggested_filename;
		if (filename == null || filename == "")
			filename = "nuvola-player-download";
		var dest_dir = File.new_for_path(dir);
		var destination = dest_dir.get_child(filename);
		var i = 0;
		while (destination.query_exists())
			destination = dest_dir.get_child("np(%d)-%s".printf(++i,filename));
		download_started(request, destination);
		message("Download: %s", destination.get_path());
		request.set_destination_uri(destination.get_uri());
		return true;
	}
	
	private void on_user_warning(string title, string text, ref bool show_again)
	{
		var dialog = new Diorite.Widgets.WarningDialog(title, text, _("Do not show warning for this service again"));
		dialog.secondary_use_markup = true;
		dialog.run();
		show_again = dialog.show_again;
	}
	
	private void on_user_question(string title, string text, ref bool response)
	{
		var dialog = new Diorite.Widgets.QuestionDialog(title, text);
		dialog.secondary_use_markup = true;
		response = (dialog.run() == Gtk.ResponseType.YES);
	}
}

} // namespace Nuvola
