/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;
using Diorite.System;

namespace Nuvola.Extensions.DockManager{

public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of a plugin providing integration with dockmanager
		_("Dock Manager"),
		Config.VERSION,
		/// Extension descriptiom
		_("<p>This plugin provides integration with docks compliant with <a href=\"https://launchpad.net/dockmanager\">DockManager specification</a> (<a href=\"http://wiki.go-docky.com\">Docky</a>, <a href=\"https://launchpad.net/awn/\">Avant Window Navigator</a>, <a href=\"https://launchpad.net/dockbar\">DockBarX</a>).</p>"),
		"Jiří Janoušek",
		typeof(Extension),
		true
	};
}

/**
 * Manages DockManager compliant docks
 */
public class Extension: Nuvola.Extension{
	private DBusDockManager dock;
	private DBusDockItem dock_item;
	private DockyDockItem docky_item;
	private Gee.Map<int32, Diorite.Action> dock_menu_items = null;
	private bool have_docky = false;
	private const ulong SLEEP = 1000;
	private bool active = false;
	private uint update_id = 0;
	
	private weak Diorite.Application app;
	private weak Diorite.Actions actions;
	
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws ExtensionError{
		// Get dependencies, throws ExtensionError
		app = objects.get<Diorite.Application>("application");
		actions = objects.get<Diorite.Actions>("actions");
		dock_menu_items = new Gee.HashMap<int32, Diorite.Action>(null, null);
		Bus.watch_name(BusType.SESSION, "net.launchpad.DockManager",
				BusNameWatcherFlags.NONE,
				dock_appeared, dock_vanished);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload(){
		debug("~DockManager");
		remove_menu();
	}
	
	private void dock_appeared(DBusConnection conn, string name, string owner){
		lock(this.active){
			this.active = false;
		}
		Idle.add(activate_dock_cb);
	}
	
	private bool activate_dock_cb(){
		
			this.dock_item = null;
			try{
				this.dock = Bus.get_proxy_sync(BusType.SESSION,
				"net.launchpad.DockManager", "/net/launchpad/DockManager");
				
				/*
				 * Capabilities are used to detect Docky, because docky isn't fully
				 * compatible with DockManager specification (wrong signature
				 * of several methods).
				 */
				var cap = this.dock.get_capabilities();
				this.have_docky = ("x-docky-get-item-pids" in cap);
				if(this.have_docky){
					debug("DockManager: Using Docky compatibility mode.");
				}
				else{
					debug("DockManager: Using standard DockManager implementation.");
				}
				cap = null;
				this.dock.item_added.connect(this.check_dock_item);
				this.dock.item_removed.connect(this.item_removed);
				foreach(string path in this.dock.get_items()){
					this.check_dock_item(path);
					if(this.dock_item != null || this.docky_item != null){
						break;
					}
				}			
			}
			catch(IOError e){
				warning("Unable to get proxy for dock: %s", e.message);
				//this.dock_cleanup();
				
			}
		
		return false;
	}
	
	
	
	/**
	 * Check if the dock item is our dock item
	 * 
	 * @param path DBus path to dock item
	 */
	private void check_dock_item(string path){
		lock(this.active){
			if(!this.have_docky){
				try{
					DBusDockItem item = Bus.get_proxy_sync(BusType.SESSION,
								"net.launchpad.DockManager", path);
					var desktop_entry = item.desktop_file;
					if(desktop_entry == null) return;
					
					if(desktop_entry.has_suffix("/" + app.desktop_entry + ".desktop")){
						debug("Dock item: %s", desktop_entry);
						this.dock_item = item;
						this.docky_item = null;
						this.create_menu();
						item.menu_item_activated.connect(this.dock_menu_item_activated);
						active = true;
					}
				}
				catch(IOError e){
					warning("Dock.Manager: Unable to get DockItem proxy for %s: %s",
					            path, e.message);
				}
			}
			else{
				try{
					DockyDockItem d_item = Bus.get_proxy_sync(BusType.SESSION,
								"net.launchpad.DockManager", path);
					var desktop_entry = d_item.desktop_file;
					if(desktop_entry == null) return;
					debug("Dock item: %s", desktop_entry);
					if(desktop_entry.has_suffix(app.desktop_entry + ".desktop")){
						this.docky_item = d_item;
						this.dock_item = null;
						this.create_menu();
						d_item.menu_item_activated.connect(this.docky_menu_item_activated);
						/* Docky requires confirmation of menu items */
						d_item.menu_item_confirmation_needed.connect(
							this.docky_menu_item_confirmation_needed);
						active = true;
					}
				}
				catch(IOError e){
					warning("Dock.Manager: Unable to get DockItem proxy for %s: %s",
					            path, e.message);
				}
			}
			if(dock_item != null || docky_item != null){
				foreach(var name in PLAYER_ACTIONS_ALL){
					var action = actions.get_action(name);
					if(action != null){
						if(name == Actions.TOGGLE_PLAY){
							action.notify.connect_after(on_actions_changed);
						}
						else{
							action.notify["sensitive"].connect_after(on_actions_changed);
						}
					}
				}
			}
		}
	}
	
	private void item_removed(string path){
		lock(active){
			active = false;
		}
	}
	
	/*
	 * Make sure this function is called inside lock(this.dock_menu_items) 
	 */
	private void create_menu(){
		debug("Create menu");
		weak DBusDockItem item = null;
		weak DockyDockItem d_item = null;
		if(this.dock_item != null){
			item = this.dock_item;
		}
		else if(this.docky_item != null){
			d_item = this.docky_item;
		}
		else{
			return;
		}
		
		
		
		var old_items = this.dock_menu_items;
		this.dock_menu_items = new Gee.HashMap<int32, Diorite.Actions>(null, null);
		this.remove_menu_items(old_items);
		try{
			GLib.HashTable<string, Variant> menu_item;
			foreach(var name in PLAYER_ACTIONS_ALL){
				var action = actions.get_action(name);
				if(action != null && action.sensitive){
					menu_item = new GLib.HashTable<string, Variant>(null, null);
//~ 						debug("label: %s → %s, %s", name, action.label ?? "null", action.stock_id ?? "null");
					menu_item.insert("label", (action.label ?? "null").replace("_", ""));
					if(action.stock_id != null){
						menu_item.insert("icon-name", action.stock_id);
					}
				
					int32 id; 
					if(item != null){
						id = item.add_menu_item(menu_item);
					}
					else{
						id = (int32) d_item.add_menu_item(menu_item);
					}
					this.dock_menu_items[id] = action;
				}
			}
		}
		catch(IOError e){
			warning("Unable to create DockManager menu: %s", e.message);
		}
	
		
	}
	
	/*
	 * Make sure this function is called inside lock(this.dock_menu_items) 
	 */
	private void remove_menu_items(Gee.Map<int32, Diorite.Action> old_items){
		try{
			foreach(var id in old_items.keys){
				Thread.usleep(SLEEP);
			
//~ 					PlayerActions action = old_items[id];
//~ 					debug("Remove DockItem: %d, %s", (int)id, action.get_id());
				if(this.dock_item != null){
					this.dock_item.remove_menu_item(id);
				}
				else if(this.docky_item != null){
					this.docky_item.remove_menu_item((uint32) id);
				}
			
			
			}
		}
		catch(IOError e){
			warning("App: Unable to remove dock menu item: %s", e.message);
		}
		old_items.clear();
	}
	
	
	private void on_actions_changed(Object o, ParamSpec p){
		debug("actions changed");
		lock(update_id){
			if(update_id == 0){
				update_id = Idle.add(update_menu_cb);
			}
		}
	}
	
	private bool update_menu_cb(){
		lock(this.active){
			if(active){
				this.update_menu();
			}
		}
		return false;
	}
	
	
	private void update_menu(){
		lock(update_id){
			create_menu();
			update_id = 0;
		}
	}
	
	
	/**
	 * Removes menu items from our dock item
	 */
	private void remove_menu(){
		lock(this.active){
			this.active = false;
		}
		this.remove_menu_items(this.dock_menu_items);
		
	}
	
	
	/**
	 * Confirms menu items at Docky
	 */
	private void docky_menu_item_confirmation_needed(){
		debug("Docky menu item confirmation");
		if(this.docky_item == null){
			warning("Docky menu item confirmation needed, but docky item is null.");
			return;
		}
		foreach(var id in this.dock_menu_items.keys){
			try{
				this.docky_item.confirm_menu_item((uint32) id);
			}
			catch(IOError e){
				warning("Unable to confirm docky menu item: %s", e.message);
			}
		}
	}
	
	/**
	 * Wrapper for DockManager.menu_item_activated signal for Docky due to
	 * signature incompatibility.
	 */
	private void docky_menu_item_activated(uint32 id){
		debug(@"Docky menu item activated: $id");
		this.dock_menu_item_activated((int32) id);
	}
	
	/**
	 * Triggers menu item action
	 */
	private void dock_menu_item_activated(int32 id){
		debug(@"Dock menu item activated: $id");
		if(this.dock_menu_items.has_key(id)){
			dock_menu_items[id].activate();
		}
	}
	
	private void dock_vanished(DBusConnection conn, string name){
		debug("Dock vanished: %s", name);
		this.dock_cleanup();
	}
	
	private void dock_cleanup(){
		lock(active){
			this.active = false;
			this.dock_item = null;
			this.docky_item = null;
			this.dock = null;
			this.dock_menu_items.clear();
		}
	}
}

[DBus(name = "net.launchpad.DockManager")]
public interface DBusDockManager: Object{
	public abstract string[] get_capabilities() throws IOError;
	public abstract string[] get_items() throws IOError;
	public abstract string[] get_items_by_desktop_file(string desktop_file) throws IOError;
	public signal void item_added(string path);
	public signal void item_removed(string path);
	
}

[DBus(name = "net.launchpad.DockItem")]
public interface DBusDockItem: Object{
	public abstract string desktop_file{ owned get;}
	public abstract int32 add_menu_item(GLib.HashTable<string, Variant> menu) throws IOError;
	public abstract void update_menu_item(GLib.HashTable<string, Variant> menu) throws IOError;
	public abstract void remove_menu_item(int32 id) throws IOError;
	public signal void menu_item_activated(int32 id);
}

[DBus(name = "net.launchpad.DockItem")]
public interface DockyDockItem: Object{
	public abstract string desktop_file{ owned get;}
	public abstract uint32 add_menu_item(GLib.HashTable<string, Variant> menu) throws IOError;
	public abstract void update_menu_item(GLib.HashTable<string, Variant> menu) throws IOError;
	public abstract void confirm_menu_item(uint32 id) throws IOError;
	public abstract void remove_menu_item(uint32 id) throws IOError;
	public signal void menu_item_activated(uint32 id);
	public signal void menu_item_confirmation_needed();
}

} // namespace Nuvola.Extensions.DockManager
