/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if NOTIFICATIONS

using Diorite.Logger;
using Diorite;
using Nuvola.Settings;

namespace Nuvola.Extensions.Notifications{

const string CONF_ID = "notifications";
const string ACTIVE_WINDOW = EXTENSIONS_KEY + "." + CONF_ID + "." + "active_window";
const string RESIDENT = EXTENSIONS_KEY + "." + CONF_ID + "." + "resident";

public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of a plugin providing integration with multimedia keys in GNOME
		_("Notifications"),
		Config.VERSION,
		/// Description of a plugin providing integration with multimedia keys in GNOME
		_("<p>This plugin provides desktop notifications (<i>libnotify</i>).</p>"),
		"Jiří Janoušek",
		typeof(Extension),
		true
	};
}

/**
 * Manages notifications
 */
public class Extension: Nuvola.Extension{
	private weak Diorite.Application app;
	private weak Player player;
	private weak Gtk.Window main_window;
	private weak MultiTypeMap config;
	private weak Diorite.Actions actions;
	private Notify.Notification? notification = null;
	private string? title = null;
	private string? message = null;
	private string? icon = null;
	private bool visible = false;
	private bool actions_supported = false;
	private bool persistence_supported = false;
	private bool icons_supported = false;
	private uint queued_id = 0;
	
	public bool active_window{
		get{
			return config.get_bool(ACTIVE_WINDOW, true);
		}
		set{
			config.set_bool(ACTIVE_WINDOW, value);
		}
	}
	
	private bool _resident = false;
	public bool resident
	{
		get
		{
			return _resident;
		}
		set
		{
			_resident = value;
			config.set_bool(RESIDENT, value);
			show_notification(true); // apply change
		}
	}
	
	construct{
		has_preferences = true;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws ExtensionError{
		// Get dependencies, throws ExtensionError
		app = objects.get<Diorite.Application>("application");
		player = objects.get<Player>("player");
		main_window = objects.get<Gtk.Window>("main_window");
		config = objects.get<Diorite.MultiTypeMap>("config");
		actions = objects.get<Diorite.Actions>("actions");
		
		Notify.init(app.display_name);
		List<string> capabilities = Notify.get_server_caps();
		persistence_supported =  capabilities.find_custom("persistence", strcmp) != null;
		actions_supported =  capabilities.find_custom("actions", strcmp) != null;
		icons_supported =  capabilities.find_custom("action-icons", strcmp) != null;
		debug(@"Notifications: persistence $persistence_supported, actions $actions_supported, icons $icons_supported");
		title = null;
		_resident = config.get_bool(RESIDENT, true);
		if (actions_supported)
		{
			foreach (var name in PLAYER_ACTIONS)
			{
				var action = actions.get_action(name);
				if (action != null)
					action.notify["visible"].connect_after(on_actions_changed);
			}
		}
				
		player.song_changed.connect(on_song_changed);
		player.notify["playback-state"].connect(on_state_changed);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload(){
		player.notify["playback-state"].disconnect(on_state_changed);
		player.song_changed.disconnect(on_song_changed);
		
		if (actions_supported)
		{
			foreach (var name in PLAYER_ACTIONS)
			{
				var action = actions.get_action(name);
				if (action != null)
					action.notify["visible"].disconnect(on_actions_changed);
			}
		}
		
		if (notification != null)
		{
			notification.closed.disconnect(on_notification_closed);
			try
			{
				notification.close();
			}
			catch(GLib.Error e){}
			visible = false;
			notification = null;
		}
		Notify.uninit();
		
		icon = null;
		title = null;
		message = null;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override Gtk.Widget? get_preferences(){
		var box = new Gtk.Grid();
		box.orientation = Gtk.Orientation.VERTICAL;
		/// Settings dialog - tab General
		var active_window = new Gtk.CheckButton.with_label(_("Notify also when window is active."));
		active_window.active = this.active_window;
		active_window.toggled.connect(() => { this.active_window = active_window.active; });
		box.add(active_window);
		
		/// Settings dialog - tab General
		var resident = new Gtk.CheckButton.with_label(_("Enable resident notifications"));
		resident.sensitive = persistence_supported;
		resident.active = this.resident && persistence_supported;
		
		resident.toggled.connect(() => { this.resident = resident.active; });
		box.add(resident);
		
		box.show_all();
		return box;
	}
	
	private void on_state_changed(Object o, ParamSpec p){
		show_notification(false);
	}
	
	private void on_song_changed(string? song, string? artist, string? album, string? album_art){
		if(artist == null && album == null){
			/// Shown in notification if both artist and album are unknown
			message = _("by unknown artist");
		}
		else if(artist == null){
			/// Shown in notification if artist is unknown, %s will be replaced by album name
			message = _("by unknown artist from %s").printf(album);
		}
		else if(album == null){
			/// Shown in notification if album is unknown, %s will be replaced by artist name
			message = _("by %s").printf(artist);
		}
		else{
			/// Shown in notification, %1$s will be replaced by artist name, %2$s will be replaced by album name
			message = _("by %1$s from %2$s").printf(artist, album);
		}
		title = song;
		icon = album_art;
		show_notification();
	}
	
	private void show_notification(bool ignore_paused=true){
		debug("Show notification");
		if (!(resident && persistence_supported) && player.playback_state != STATE_PLAYING
		&& ignore_paused)
			return;
		lock(queued_id){
			if(queued_id == 0){
				queued_id = Idle.add(show_notification_cb);
			}
		}
	}
	
	private bool show_notification_cb(){
		debug("Show notification callback");
		lock(queued_id){
			queued_id = 0;
		}
		
		if (title == null || (main_window.is_active && !active_window))
			return false;
		try{
			if (notification == null){
				notification = new Notify.Notification(title, message, ""); 
			}
			else{
				notification.clear_hints();
				notification.clear_actions();
				notification.update(title, message, "");
			}
			
			if (icon != null)
			{
				try
				{
					// Pass actual image data over dbus instead of a filename to
					// prevent caching. LP:1099825
					notification.set_image_from_pixbuf(new Gdk.Pixbuf.from_file(icon));
				}
				catch(GLib.Error e)
				{
					warning("Failed to load album art %s: %s", icon, e.message);
				}
			}
			
			notification.set_category("x-gnome.music");
			notification.set_hint("desktop-entry", app.desktop_entry);
			
			if (actions_supported)
			{
				if (icons_supported)
					notification.set_hint("action-icons", true);
				
				foreach (var name in PLAYER_ACTIONS)
				{
					var action = actions.get_action(name);
					if (action != null && action.sensitive)
					{
						notification.add_action(action.icon_name ?? action.stock_id, action.label,
						() => { action.activate(); });
					}
				}
			}
			
			if (persistence_supported && resident)
				notification.set_hint("resident", true);
			
			notification.show();
			visible = true;
		}
		catch(Error e){
			warning("Unable to show notification: %s", e.message);
		}
		
		return false;
	}
	
	private void on_notification_closed(){
		debug("Notification closed");
		visible = false;
	}
	
	private void on_actions_changed(Object o, ParamSpec p){
		debug("Actions changed.");
		show_notification(false);
	}
}

} // namespace Nuvola.Extensions.Notifications
#endif
